"""
versionfinder/tests/test_versioninfo.py

The latest version of this package is available at:
<https://github.com/jantman/versionfinder>

################################################################################
Copyright 2015 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of versionfinder.

    versionfinder is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    versionfinder is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with versionfinder.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the GPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/versionfinder> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import sys
from versionfinder.versioninfo import VersionInfo

# https://code.google.com/p/mock/issues/detail?id=249
# py>=3.4 should use unittest.mock not the mock package on pypi
if (
        sys.version_info[0] < 3 or
        sys.version_info[0] == 3 and sys.version_info[1] < 4
):
    from mock import patch, PropertyMock
else:
    from unittest.mock import patch, PropertyMock

pb = 'versionfinder.versioninfo.VersionInfo'


class TestInit(object):

    def test_init_all(self):
        v = VersionInfo(
            pip_version='pipver',
            pip_url='pipurl',
            pip_requirement='preq',
            pkg_resources_version='prver',
            pkg_resources_url='prurl',
            git_tag='tag',
            git_commit='commit',
            git_remotes={
                'origin': 'ourl'
            },
            git_is_dirty=True
        )
        assert v._pip_version == 'pipver'
        assert v._pip_url == 'pipurl'
        assert v._pip_requirement == 'preq'
        assert v._pkg_resources_version == 'prver'
        assert v._pkg_resources_url == 'prurl'
        assert v._git_tag == 'tag'
        assert v._git_commit == 'commit'
        assert v._git_remotes == {
            'origin': 'ourl'
        }
        assert v._git_is_dirty is True

    def test_init_empty(self):
        v = VersionInfo()
        assert v._pip_version is None
        assert v._pip_url is None
        assert v._pip_requirement is None
        assert v._pkg_resources_version is None
        assert v._pkg_resources_url is None
        assert v._git_tag is None
        assert v._git_commit is None
        assert v._git_remotes is None
        assert v._git_is_dirty is None


class TestAsDict(object):

    def test_dict(self):
        d = {
            'pip_version': 'pipver',
            'pip_url': 'pipurl',
            'pip_requirement': 'preq',
            'pkg_resources_version': 'prver',
            'pkg_resources_url': 'prurl',
            'git_tag': 'tag',
            'git_commit': 'commit',
            'git_remotes': {
                'origin': 'ourl'
            },
            'git_is_dirty': True
        }
        v = VersionInfo(**d)
        assert v.as_dict == d


class TestProperties(object):

    def setup_method(self, _):
        self.cls = VersionInfo(
            pip_version='pipver',
            pip_url='pipurl',
            pip_requirement='preq',
            pkg_resources_version='prver',
            pkg_resources_url='prurl',
            git_tag='tag',
            git_commit='commit',
            git_remotes={
                'origin': 'ourl'
            },
            git_is_dirty=True
        )

    def test_version(self):
        assert self.cls.version == 'pipver'

    def test_version_no_pip(self):
        self.cls._pip_version = None
        assert self.cls.version == 'prver'

    def test_url(self):
        assert self.cls.url == 'pipurl'

    def test_url_no_pip(self):
        self.cls._pip_url = None
        assert self.cls.url == 'prurl'

    def test_pip_version(self):
        assert self.cls.pip_version == 'pipver'

    def test_pip_url(self):
        assert self.cls.pip_url == 'pipurl'

    def test_pip_requirement(self):
        assert self.cls.pip_requirement == 'preq'

    def test_pkg_resources_version(self):
        assert self.cls.pkg_resources_version == 'prver'

    def test_pkg_resources_url(self):
        assert self.cls.pkg_resources_url == 'prurl'

    def test_git_tag(self):
        assert self.cls.git_tag == 'tag'

    def test_git_commit(self):
        assert self.cls.git_commit == 'commit'

    def test_git_remotes(self):
        assert self.cls.git_remotes == {
            'origin': 'ourl'
        }

    def test_git_remote(self):
        assert self.cls.git_remote == 'ourl'

    def test_git_remote_none(self):
        self.cls._git_remotes = None
        assert self.cls.git_remote is None

    def test_git_remote_no_origin(self):
        self.cls._git_remotes = {
            'a': 'rmta',
            'k': 'rmtk',
            'z': 'rmtz'
        }
        assert self.cls.git_remote == 'rmta'

    def test_git_is_dirty(self):
        assert self.cls.git_is_dirty is True

    def test_git_str(self):
        assert self.cls.git_str == 'ourl@tag*'

    def test_git_str_no_git(self):
        self.cls._git_commit = None
        self.cls._git_remotes = None
        assert self.cls.git_str == ''

    def test_git_str_no_tag(self):
        self.cls._git_tag = None
        assert self.cls.git_str == 'ourl@commit*'

    def test_git_str_not_dirty(self):
        self.cls._git_is_dirty = False
        assert self.cls.git_str == 'ourl@tag'

    def test_git_str_pip_req_not_dirty(self):
        self.cls._git_is_dirty = False
        self.cls._pip_requirement = 'git+https://foo'
        assert self.cls.git_str == 'git+https://foo'

    def test_git_str_pip_req_dirty(self):
        self.cls._pip_requirement = 'git+https://foo'
        assert self.cls.git_str == 'git+https://foo*'

    def test_short_str_pip(self):
        assert self.cls.short_str == 'pipver <pipurl>'

    def test_short_str_pkg_resources(self):
        self.cls._pip_version = None
        self.cls._pip_url = None
        assert self.cls.short_str == 'prver <prurl>'

    def test_short_str_none(self):
        with patch('%s.version' % pb, new_callable=PropertyMock) as m_ver:
            with patch('%s.url' % pb, new_callable=PropertyMock) as m_url:
                m_ver.return_value = None
                m_url.return_value = None
                assert self.cls.short_str == ''

    def test_long_str_git(self):
        self.cls._git_is_dirty = False
        with patch('%s.git_str' % pb, new_callable=PropertyMock) as m_git_str:
            with patch('%s.short_str' % pb,
                       new_callable=PropertyMock) as m_short:
                m_git_str.return_value = 'gitstr'
                m_short.return_value = 'shortstr'
                assert self.cls.long_str == 'shortstr (gitstr)'

    def test_long_str_no_git(self):
        with patch('%s.git_str' % pb, new_callable=PropertyMock) as m_git_str:
            with patch('%s.short_str' % pb,
                       new_callable=PropertyMock) as m_short:
                m_git_str.return_value = ''
                m_short.return_value = 'shortstr'
                assert self.cls.long_str == 'shortstr'


class TestRepr(object):

    def test_repr(self):
        v = VersionInfo(
            pip_version='pipver',
            pip_url='pipurl',
            pip_requirement='preq',
            pkg_resources_version='prver',
            pkg_resources_url='prurl',
            git_tag='tag',
            git_commit='commit',
            git_remotes={
                'origin': 'ourl'
            },
            git_is_dirty=True
        )
        s = 'VersionInfo('
        s += 'git_commit=commit, '
        s += 'git_is_dirty=True, '
        s += "git_remotes={'origin': 'ourl'}, "
        s += 'git_tag=tag, '
        s += 'pip_requirement=preq, '
        s += 'pip_url=pipurl, '
        s += 'pip_version=pipver, '
        s += 'pkg_resources_url=prurl, '
        s += 'pkg_resources_version=prver'
        s += ')'
        assert v.__repr__() == s


class TestEq(object):

    def test_equal(self):
        v1 = VersionInfo(
            pip_version='pipver',
            pip_url='pipurl',
            pip_requirement='preq',
            pkg_resources_version='prver',
            pkg_resources_url='prurl',
            git_tag='tag',
            git_commit='commit',
            git_remotes={
                'origin': 'ourl'
            },
            git_is_dirty=True
        )
        v2 = VersionInfo(
            pip_version='pipver',
            pip_url='pipurl',
            pip_requirement='preq',
            pkg_resources_version='prver',
            pkg_resources_url='prurl',
            git_tag='tag',
            git_commit='commit',
            git_remotes={
                'origin': 'ourl'
            },
            git_is_dirty=True
        )
        assert v1 == v2

    def test_not_equal(self):
        v1 = VersionInfo(
            pip_version='pipver',
            pip_url='pipurl',
            pip_requirement='preq',
            pkg_resources_version='prver',
            pkg_resources_url='prurl',
            git_tag='tag',
            git_commit='commit',
            git_remotes={
                'origin': 'ourl'
            },
            git_is_dirty=True
        )
        v2 = VersionInfo(
            pip_version='pipver',
            pip_url='pipurl',
            pip_requirement='preq',
            pkg_resources_version='prver',
            pkg_resources_url='prurl',
            git_tag='tag',
            git_commit='commit',
            git_remotes={
                'origin': 'ourl'
            },
            git_is_dirty=False
        )
        v3 = VersionInfo(
            pip_version='pipver',
            pip_url='pipurl',
            pip_requirement='preq',
            pkg_resources_version='prver',
            pkg_resources_url='prurl',
            git_tag='tag',
            git_commit='commit',
            git_remotes={
                'origin': 'ourl2'
            },
            git_is_dirty=True
        )
        assert v1 != v2
        assert v1 != v3



"""
versionfinder/tests/test_versioninfo.py

The latest version of this package is available at:
<https://github.com/jantman/versionfinder>

################################################################################
Copyright 2015 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of versionfinder.

    versionfinder is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    versionfinder is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with versionfinder.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the GPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/versionfinder> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import sys
from versionfinder import find_version

# https://code.google.com/p/mock/issues/detail?id=249
# py>=3.4 should use unittest.mock not the mock package on pypi
if (
        sys.version_info[0] < 3 or
        sys.version_info[0] == 3 and sys.version_info[1] < 4
):
    from mock import patch, call, Mock
else:
    from unittest.mock import patch, call, Mock


class TestFindVersion(object):

    def test_caller_passed(self):
        m_frame = Mock()
        m_result = Mock()
        with patch('versionfinder.VersionFinder', autospec=True) as mock_vf:
            with patch('versionfinder.inspect.stack') as mock_stack:
                mock_vf.return_value.find_package_version.\
                    return_value = m_result
                res = find_version('pname', caller_frame=m_frame)
        assert mock_vf.mock_calls == [
            call('pname', caller_frame=m_frame),
            call().find_package_version()
        ]
        assert mock_stack.mock_calls == []
        assert res == m_result

    def test_caller_not_passed(self):
        m_frame = Mock()
        m_result = Mock()
        with patch('versionfinder.VersionFinder', autospec=True) as mock_vf:
            with patch('versionfinder.inspect.stack') as mock_stack:
                mock_vf.return_value.find_package_version.\
                    return_value = m_result
                mock_stack.return_value = [None, [m_frame]]
                res = find_version('pname')
        assert mock_vf.mock_calls == [
            call('pname', caller_frame=m_frame),
            call().find_package_version()
        ]
        assert mock_stack.mock_calls == [call()]
        assert res == m_result



"""
versionfinder/versionfinder.py

The latest version of this package is available at:
<https://github.com/jantman/versionfinder>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of versionfinder.

    versionfinder is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    versionfinder is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with versionfinder.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the GPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/versionfinder> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import os
import logging
import inspect
from contextlib import contextmanager

from .versioninfo import VersionInfo

try:
    import pip._internal as pip
except (ImportError, KeyError):
    try:
        import pip
    except (ImportError, KeyError):
        # this is used within try blocks; NBD if they fail
        pass

try:
    import pkg_resources
except ImportError:
    # this is used within try blocks; NBD if they fail
    pass

try:
    from git import Repo
except Exception:
    # this is used within try blocks; NBD if they fail
    pass

logger = logging.getLogger(__name__)


class VersionFinder(object):

    def __init__(self, package_name, package_file=None, log=False,
                 caller_frame=None):
        """
        Initialize a VersionFinder to find version information of the named
        package, which includes a given file. ``package_file`` must be a Python
        file in the package; if not specified, the file calling this class
        will be used.

        VersionFinder logs rather verbosely to ``logging.debug()`` if ``log``
        is True. To simplify use as a library, unless you set ``log`` to True,
        versionfinder's logger will be set to a level of ``logging.CRITICAL``,
        suppressing all log messages. This will also silence the ``pip`` logger.

        :param package_name: name of the package to find information about
        :type package_name: str
        :param package_file: absolute path to a Python source file in the
          package to find information about; if not specified, the file calling
          this class will be used
        :type package_file: str
        :param log: If not set to True, the "versionfinder" and "pip" loggers
          will be set to a level of :py:const:`logging.CRITICAL` to suppress
          log output. If set to True, you will see a LOT of debug-level log
          output, for debugging the internals of versionfinder.
        :type log: bool
        :param caller_frame: If the call to this method is wrapped by something
          else, this should be the stack frame representing the original caller.
          Not used if ``package_file`` is specified. See
          :py:func:`versionfinder.find_version` for an example.
        :type caller_frame: frame
        """
        if not log:
            logger.setLevel(logging.CRITICAL)
            pip_log = logging.getLogger("pip")
            pip_log.setLevel(logging.CRITICAL)
            pip_log.propagate = True
        logger.debug("Finding package version for: %s", package_name)
        self.package_name = package_name
        if package_file is not None:
            logger.debug("Explicit package file: %s", package_file)
            self.package_file = package_file
        else:
            if caller_frame is None:
                caller_frame = inspect.stack()[1][0]
            self.package_file = os.path.abspath(
                inspect.getframeinfo(caller_frame).filename)
            logger.debug("Found package_file as: %s", self.package_file)
        self.package_dir = os.path.dirname(self.package_file)
        logger.debug('package_dir: %s' % self.package_dir)
        self._pip_locations = []
        self._pkg_resources_locations = []

    def find_package_version(self):
        """
        Find the installed version of the specified package, and as much
        information about it as possible (source URL, git ref or tag, etc.)

        This attempts, to the best of our ability, to find out if the package
        was installed from git, and if so, provide information on the origin
        of that git repository and status of the clone. Otherwise, it uses
        pip and pkg_resources to find the version and homepage of the installed
        distribution.

        This class is not a sure-fire method of identifying the source of
        the distribution or ensuring AGPL compliance; it simply helps with this
        process _iff_ a modified version is installed from an editable git URL
        _and_ all changes are pushed up to the publicly-visible origin.

        Returns a dict with keys 'version', 'tag', 'commit', and 'url'.
        Values are strings or None.

        :param package_name: name of the package to find information for
        :type package_name: str
        :returns: information about the installed version of the package
        :rtype: :py:class:`~versionfinder.versioninfo.VersionInfo`
        """
        res = {
            'pip_version': None,
            'pip_url': None,
            'pip_requirement': None,
            'pkg_resources_version': None,
            'pkg_resources_url': None,
            'git_tag': None,
            'git_commit': None,
            'git_remotes': None,
            'git_is_dirty': None
        }
        try:
            pip_info = self._find_pip_info()
        except Exception:
            # we NEVER want this to crash the program
            logger.debug('Caught exception running _find_pip_info()')
            pip_info = {}
        logger.debug("pip info: %s", pip_info)
        for k, v in pip_info.items():
            if v is not None:
                res['pip_' + k] = v
        try:
            pkg_info = self._find_pkg_info()
        except Exception:
            logger.debug('Caught exception running _find_pkg_info()')
            pkg_info = {}
        logger.debug("pkg_resources info: %s", pkg_info)
        for k, v in pkg_info.items():
            res['pkg_resources_' + k] = v
        gitdir = self._git_repo_path
        if gitdir is not None:
            git_info = self._find_git_info(gitdir)
            logger.debug("Git info: %s", git_info)
            for k, v in git_info.items():
                if k == 'dirty':
                    res['git_is_dirty'] = v
                elif k == 'commit':
                    res['git_commit'] = v
                elif k == 'remotes':
                    res['git_remotes'] = v
                elif k == 'tag':
                    res['git_tag'] = v
        else:
            logger.debug("Install does not appear to be a git clone")
        logger.debug("Final package info: %s", res)
        return VersionInfo(**res)

    @property
    def _git_repo_path(self):
        """
        Attempt to determine whether this package is installed via git or not;
        if so, return the path to the git repository.

        :rtype: str
        :returns: path to git repo, or None
        """
        logger.debug('Checking for git directory in: %s', self._package_top_dir)
        for p in self._package_top_dir:
            gitdir = os.path.join(p, '.git')
            if os.path.exists(gitdir):
                logger.debug('_is_git_clone() true based on %s' % gitdir)
                return gitdir
        logger.debug('_is_git_clone() false')
        return None

    def _find_pkg_info(self):
        """
        Find information about the installed package from pkg_resources.

        :returns: information from pkg_resources about ``self.package_name``
        :rtype: dict
        """
        dist = pkg_resources.require(self.package_name)[0]
        self._pkg_resources_locations = [dist.location]
        ver, url = self._dist_version_url(dist)
        return {'version': ver, 'url': url}

    def _find_pip_info(self):
        """
        Try to find information about the installed package from pip.
        This should be wrapped in a try/except.

        :returns: information from pip about ``self.package_name``.
        :rtype: dict
        """
        res = {}
        dist = None
        dist_name = self.package_name.replace('_', '-')
        logger.debug('Checking for pip distribution named: %s', dist_name)
        for d in pip.get_installed_distributions():
            if d.project_name == dist_name:
                dist = d
        if dist is None:
            logger.debug('could not find dist matching package_name')
            return res
        logger.debug('found dist: %s', dist)
        self._pip_locations = [dist.location]
        ver, url = self._dist_version_url(dist)
        res['version'] = ver
        res['url'] = url
        # this is a bit of an ugly, lazy hack...
        req = pip.FrozenRequirement.from_dist(dist, [])
        logger.debug('pip FrozenRequirement: %s', req)
        res['requirement'] = str(req.req)
        return res

    def _dist_version_url(self, dist):
        """
        Get version and homepage for a pkg_resources.Distribution

        :param dist: the pkg_resources.Distribution to get information for
        :returns: 2-tuple of (version, homepage URL)
        :rtype: tuple
        """
        ver = str(dist.version)
        url = None
        for line in dist.get_metadata_lines(dist.PKG_INFO):
            line = line.strip()
            if ':' not in line:
                continue
            (k, v) = line.split(':', 1)
            if k == 'Home-page':
                url = v.strip()
        return (ver, url)

    def _find_git_info(self, gitdir):
        """
        Find information about the git repository, if this file is in a clone.

        :param gitdir: path to the git repo's .git directory
        :type gitdir: str
        :returns: information about the git clone
        :rtype: dict
        """
        res = {'remotes': None, 'tag': None, 'commit': None, 'dirty': None}
        try:
            logger.debug('opening %s as git.Repo', gitdir)
            repo = Repo(path=gitdir, search_parent_directories=False)
            res['commit'] = repo.head.commit.hexsha
            res['dirty'] = repo.is_dirty(untracked_files=True)
            res['remotes'] = {}
            for rmt in repo.remotes:
                # each is a git.Remote
                urls = [u for u in rmt.urls]  # generator
                if len(urls) > 0:
                    res['remotes'][rmt.name] = urls[0]
            for tag in repo.tags:
                # each is a git.Tag object
                if tag.commit.hexsha == res['commit']:
                    res['tag'] = tag.name
        except Exception:
            logger.debug('Exception getting git information', exc_info=True)
        return res

    @property
    def _package_top_dir(self):
        """
        Find one or more directories that we think may be the top-level
        directory of the package; return a list of their absolute paths.

        :return: list of possible package top-level directories (absolute paths)
        :rtype: list
        """
        r = [self.package_dir]
        for l in self._pip_locations:
            if l is not None:
                r.append(l)
        for l in self._pkg_resources_locations:
            if l is not None:
                r.append(l)
        return sorted(list(set(r)))


@contextmanager
def chdir(path):
    old_dir = os.getcwd()
    logger.debug('with chdir(%s) from %s', path, old_dir)
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_dir)



"""
versionfinder/tests/test_versionfinder.py

The latest version of this package is available at:
<https://github.com/jantman/versionfinder>

################################################################################
Copyright 2015 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of versionfinder.

    versionfinder is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    versionfinder is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with versionfinder.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the GPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/versionfinder> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
import sys
from pip._vendor.packaging.version import Version
from git import Repo

from versionfinder.versionfinder import (VersionFinder, chdir)
from versionfinder.versioninfo import VersionInfo

import logging
logger = logging.getLogger(__name__)

# https://code.google.com/p/mock/issues/detail?id=249
# py>=3.4 should use unittest.mock not the mock package on pypi
if (
        sys.version_info[0] < 3 or
        sys.version_info[0] == 3 and sys.version_info[1] < 4
):
    from mock import patch, call, DEFAULT, Mock, PropertyMock, MagicMock
else:
    from unittest.mock import (
        patch, call, DEFAULT, Mock, PropertyMock, MagicMock)

pbm = 'versionfinder.versionfinder'
pb = '%s.VersionFinder' % pbm


def mockrepo(commit=None, dirty=False, remotes={}, tag=None):
    """return a mock Repo"""
    m = MagicMock(spec_set=Repo)
    m.is_dirty.return_value = dirty
    m.head.commit.hexsha = commit
    tag1 = MagicMock()
    tag1.name = 'wrongTagName'
    tag1.commit.hexsha = '1234'
    tags = [tag1]
    if tag is not None:
        tag2 = MagicMock()
        tag2.name = tag
        tag2.commit.hexsha = commit
        tags.append(tag2)
    m.tags = tags
    rmts = []
    for name, url in remotes.items():
        rmt = MagicMock()
        rmt.name = name
        if isinstance(url, type([])):
            rmt.urls = url
        else:
            rmt.urls = [url]
        rmts.append(rmt)
    m.remotes = rmts
    return m


class BaseTest(object):

    def setup_method(self, _):
        self.cls = VersionFinder('foo', package_file='/foo/bar/baz.py')


class TestInit(object):

    def test_init(self):
        m_pip_logger = Mock()
        with patch('%s.inspect.stack' % pbm, autospec=True) as m_stack:
            with patch('%s.logger' % pbm, autospec=True) as m_logger:
                with patch('%s.logging.getLogger' % pbm,
                           autospec=True) as m_log_pip:
                    m_log_pip.return_value = m_pip_logger
                    cls = VersionFinder('foobar',
                                        package_file='/foo/bar/baz.py')
        assert m_stack.mock_calls == []
        assert cls.package_name == 'foobar'
        assert cls.package_file == '/foo/bar/baz.py'
        assert cls.package_dir == '/foo/bar'
        assert m_logger.mock_calls == [
            call.setLevel(logging.CRITICAL),
            call.debug('Finding package version for: %s', 'foobar'),
            call.debug('Explicit package file: %s', '/foo/bar/baz.py'),
            call.debug('package_dir: /foo/bar')
        ]
        assert m_log_pip.mock_calls == [call('pip')]
        assert m_pip_logger.mock_calls == [call.setLevel(logging.CRITICAL)]

    def test_init_log_true(self):
        m_pip_logger = Mock()
        with patch('%s.inspect.stack' % pbm, autospec=True) as m_stack:
            with patch('%s.logger' % pbm, autospec=True) as m_logger:
                with patch('%s.logging.getLogger' % pbm,
                           autospec=True) as m_log_pip:
                    m_log_pip.return_value = m_pip_logger
                    cls = VersionFinder('foobar', log=True,
                                        package_file='/foo/bar/baz.py')
        assert m_stack.mock_calls == []
        assert cls.package_name == 'foobar'
        assert cls.package_file == '/foo/bar/baz.py'
        assert cls.package_dir == '/foo/bar'
        assert m_logger.mock_calls == [
            call.debug('Finding package version for: %s', 'foobar'),
            call.debug('Explicit package file: %s', '/foo/bar/baz.py'),
            call.debug('package_dir: /foo/bar')
        ]
        assert m_log_pip.mock_calls == []
        assert m_pip_logger.mock_calls == []

    def test_init_no_file_no_frame(self):
        m_frame = Mock()
        type(m_frame).filename = '/tmp/foo.py'
        m_stack_frame = Mock()
        with patch('%s.inspect.stack' % pbm, autospec=True) as m_stack:
            with patch('%s.inspect.getframeinfo' % pbm) as m_get_frame:
                m_stack.return_value = [None, [m_stack_frame]]
                m_get_frame.return_value = m_frame
                cls = VersionFinder('foobar')
        assert m_stack.mock_calls == [call()]
        assert m_get_frame.mock_calls == [call(m_stack_frame)]
        assert cls.package_name == 'foobar'
        assert cls.package_file == '/tmp/foo.py'
        assert cls.package_dir == '/tmp'

    def test_init_no_file(self):
        m_frame = Mock()
        type(m_frame).filename = '/tmp/foo.py'
        m_stack_frame = Mock()
        with patch('%s.inspect.stack' % pbm, autospec=True) as m_stack:
            with patch('%s.inspect.getframeinfo' % pbm) as m_get_frame:
                m_stack.return_value = [None, []]
                m_get_frame.return_value = m_frame
                cls = VersionFinder('foobar', caller_frame=m_stack_frame)
        assert m_stack.mock_calls == []
        assert m_get_frame.mock_calls == [call(m_stack_frame)]
        assert cls.package_name == 'foobar'
        assert cls.package_file == '/tmp/foo.py'
        assert cls.package_dir == '/tmp'


class TestFindPackageVersion(BaseTest):

    def test_git_notag(self):
        with patch.multiple(
                pb,
                autospec=True,
                _find_git_info=DEFAULT,
                _find_pip_info=DEFAULT,
                _find_pkg_info=DEFAULT,
        ) as mocks:
            mocks['_find_git_info'].return_value = {
                'remotes': {'origin': 'git+https://foo'},
                'tag': None,
                'commit': '12345678',
                'dirty': False
            }
            mocks['_find_pip_info'].return_value = {
                'version': '1.2.3',
                'url': 'http://my.package.url/pip',
                'foo': None
            }
            mocks['_find_pkg_info'].return_value = {
                'version': '1.2.3',
                'url': 'http://my.package.url/pkg_resources'
            }
            with patch('%s._git_repo_path' % pb,
                       new_callable=PropertyMock) as mock_is_git:
                mock_is_git.return_value = '/git/repo/.git'
                res = self.cls.find_package_version()
        assert res.as_dict == VersionInfo(
            pip_version='1.2.3',
            pip_url='http://my.package.url/pip',
            pkg_resources_version='1.2.3',
            pkg_resources_url='http://my.package.url/pkg_resources',
            git_commit='12345678',
            git_is_dirty=False,
            git_remotes={'origin': 'git+https://foo'}
        ).as_dict
        assert mocks['_find_git_info'].mock_calls == [
            call(self.cls, '/git/repo/.git')
        ]
        assert mocks['_find_pip_info'].mock_calls == [call(self.cls)]
        assert mocks['_find_pkg_info'].mock_calls == [call(self.cls)]
        assert mock_is_git.mock_calls == [call()]

    def test_git_notag_dirty(self):
        with patch.multiple(
                pb,
                autospec=True,
                _find_git_info=DEFAULT,
                _find_pip_info=DEFAULT,
                _find_pkg_info=DEFAULT,
        ) as mocks:
            mocks['_find_git_info'].return_value = {
                'remotes': {'origin': 'git+https://foo'},
                'tag': None,
                'commit': '12345678',
                'dirty': True
            }
            mocks['_find_pip_info'].return_value = {
                'version': '1.2.3',
                'url': 'http://my.package.url/pip'
            }
            mocks['_find_pkg_info'].return_value = {
                'version': '1.2.3',
                'url': 'http://my.package.url/pkg_resources'
            }
            with patch('%s._git_repo_path' % pb,
                       new_callable=PropertyMock) as mock_is_git:
                mock_is_git.return_value = '/git/repo/.git'
                res = self.cls.find_package_version()
        assert res.as_dict == VersionInfo(
            pip_version='1.2.3',
            pip_url='http://my.package.url/pip',
            pkg_resources_version='1.2.3',
            pkg_resources_url='http://my.package.url/pkg_resources',
            git_commit='12345678',
            git_is_dirty=True,
            git_remotes={'origin': 'git+https://foo'}
        ).as_dict
        assert mocks['_find_git_info'].mock_calls == [
            call(self.cls, '/git/repo/.git')
        ]
        assert mocks['_find_pip_info'].mock_calls == [call(self.cls)]
        assert mocks['_find_pkg_info'].mock_calls == [call(self.cls)]
        assert mock_is_git.mock_calls == [call()]

    def test_git_tag(self):
        with patch.multiple(
                pb,
                autospec=True,
                _find_git_info=DEFAULT,
                _find_pip_info=DEFAULT,
                _find_pkg_info=DEFAULT,
        ) as mocks:
            mocks['_find_git_info'].return_value = {
                'remotes': {'origin': 'git+https://foo'},
                'tag': 'mytag',
                'commit': '12345678',
                'dirty': False,
                'foo': None
            }
            mocks['_find_pip_info'].return_value = {
                'version': '1.2.3',
                'url': 'http://my.package.url/pip'
            }
            mocks['_find_pkg_info'].return_value = {
                'version': '1.2.3',
                'url': 'http://my.package.url/pkg_resources'
            }
            with patch('%s._git_repo_path' % pb,
                       new_callable=PropertyMock) as mock_is_git:
                mock_is_git.return_value = '/git/repo/.git'
                res = self.cls.find_package_version()
        assert res.as_dict == VersionInfo(
            pip_version='1.2.3',
            pip_url='http://my.package.url/pip',
            pkg_resources_version='1.2.3',
            pkg_resources_url='http://my.package.url/pkg_resources',
            git_commit='12345678',
            git_is_dirty=False,
            git_tag='mytag',
            git_remotes={'origin': 'git+https://foo'}
        ).as_dict
        assert mocks['_find_git_info'].mock_calls == [
            call(self.cls, '/git/repo/.git')
        ]
        assert mocks['_find_pip_info'].mock_calls == [call(self.cls)]
        assert mocks['_find_pkg_info'].mock_calls == [call(self.cls)]
        assert mock_is_git.mock_calls == [call()]

    def test_git_tag_dirty(self):
        with patch.multiple(
                pb,
                autospec=True,
                _find_git_info=DEFAULT,
                _find_pip_info=DEFAULT,
                _find_pkg_info=DEFAULT,
        ) as mocks:
            mocks['_find_git_info'].return_value = {
                'remotes': {'origin': 'git+https://foo'},
                'tag': 'mytag',
                'commit': '12345678',
                'dirty': True
            }
            mocks['_find_pip_info'].return_value = {
                'version': '1.2.3',
                'url': 'http://my.package.url/pip'
            }
            mocks['_find_pkg_info'].return_value = {
                'version': '1.2.3',
                'url': 'http://my.package.url/pkg_resources'
            }
            with patch('%s._git_repo_path' % pb,
                       new_callable=PropertyMock) as mock_is_git:
                mock_is_git.return_value = '/git/repo/.git'
                res = self.cls.find_package_version()
        assert res.as_dict == VersionInfo(
            pip_version='1.2.3',
            pip_url='http://my.package.url/pip',
            pkg_resources_version='1.2.3',
            pkg_resources_url='http://my.package.url/pkg_resources',
            git_commit='12345678',
            git_is_dirty=True,
            git_tag='mytag',
            git_remotes={'origin': 'git+https://foo'}
        ).as_dict
        assert mocks['_find_git_info'].mock_calls == [
            call(self.cls, '/git/repo/.git')
        ]
        assert mocks['_find_pip_info'].mock_calls == [call(self.cls)]
        assert mocks['_find_pkg_info'].mock_calls == [call(self.cls)]
        assert mock_is_git.mock_calls == [call()]

    def test_pkg_res_exception(self):

        def se_exception():
            raise Exception("some exception")

        with patch.multiple(
                pb,
                autospec=True,
                _find_git_info=DEFAULT,
                _find_pip_info=DEFAULT,
                _find_pkg_info=DEFAULT,
        ) as mocks:
            mocks['_find_git_info'].return_value = {
                'remotes': {'origin': 'git+https://foo'},
                'tag': None,
                'commit': '12345678',
                'dirty': False
            }
            mocks['_find_pip_info'].return_value = {
                'version': '1.2.3',
                'url': 'http://my.package.url/pip'
            }
            mocks['_find_pkg_info'].side_effect = se_exception
            with patch('%s._git_repo_path' % pb,
                       new_callable=PropertyMock) as mock_is_git:
                mock_is_git.return_value = '/git/repo/.git'
                res = self.cls.find_package_version()
        assert res.as_dict == VersionInfo(
            pip_version='1.2.3',
            pip_url='http://my.package.url/pip',
            git_commit='12345678',
            git_is_dirty=False,
            git_remotes={'origin': 'git+https://foo'}
        ).as_dict
        assert mocks['_find_git_info'].mock_calls == [
            call(self.cls, '/git/repo/.git')
        ]
        assert mocks['_find_pip_info'].mock_calls == [call(self.cls)]
        assert mocks['_find_pkg_info'].mock_calls == [call(self.cls)]
        assert mock_is_git.mock_calls == [call()]

    def test_pip_exception(self):

        def se_exception():
            raise Exception("some exception")

        with patch.multiple(
                pb,
                autospec=True,
                _find_git_info=DEFAULT,
                _find_pip_info=DEFAULT,
                _find_pkg_info=DEFAULT,
        ) as mocks:
            mocks['_find_git_info'].return_value = {
                'remotes': {'origin': 'git+https://foo'},
                'tag': None,
                'commit': '12345678',
                'dirty': False
            }
            mocks['_find_pip_info'].side_effect = se_exception
            mocks['_find_pkg_info'].return_value = {
                'version': '1.2.3',
                'url': 'http://my.package.url/pkg_resources'
            }
            with patch('%s._git_repo_path' % pb,
                       new_callable=PropertyMock) as mock_is_git:
                mock_is_git.return_value = '/git/repo/.git'
                res = self.cls.find_package_version()
        assert res.as_dict == VersionInfo(
            pkg_resources_version='1.2.3',
            pkg_resources_url='http://my.package.url/pkg_resources',
            git_commit='12345678',
            git_is_dirty=False,
            git_remotes={'origin': 'git+https://foo'}
        ).as_dict
        assert mocks['_find_git_info'].mock_calls == [
            call(self.cls, '/git/repo/.git')
        ]
        assert mocks['_find_pip_info'].mock_calls == [call(self.cls)]
        assert mocks['_find_pkg_info'].mock_calls == [call(self.cls)]
        assert mock_is_git.mock_calls == [call()]

    def test_no_git(self):

        with patch.multiple(
                pb,
                autospec=True,
                _find_git_info=DEFAULT,
                _find_pip_info=DEFAULT,
                _find_pkg_info=DEFAULT,
        ) as mocks:
            mocks['_find_git_info'].return_value = {
                'remotes': {},
                'tag': None,
                'commit': None,
                'dirty': None,
            }
            mocks['_find_pip_info'].return_value = {
                'version': '1.2.3',
                'url': 'http://my.package.url/pip'
            }
            mocks['_find_pkg_info'].return_value = {
                'version': '1.2.3',
                'url': 'http://my.package.url/pkg_resources'
            }
            with patch('%s._git_repo_path' % pb,
                       new_callable=PropertyMock) as mock_is_git:
                mock_is_git.return_value = None
                res = self.cls.find_package_version()
        assert res.as_dict == VersionInfo(
            pip_version='1.2.3',
            pip_url='http://my.package.url/pip',
            pkg_resources_version='1.2.3',
            pkg_resources_url='http://my.package.url/pkg_resources',
        ).as_dict
        assert mocks['_find_git_info'].mock_calls == []
        assert mocks['_find_pip_info'].mock_calls == [call(self.cls)]
        assert mocks['_find_pkg_info'].mock_calls == [call(self.cls)]
        assert mock_is_git.mock_calls == [call()]

    def test_no_git_no_pip(self):

        def se_exception():
            raise Exception("some exception")

        with patch.multiple(
                pb,
                autospec=True,
                _find_git_info=DEFAULT,
                _find_pip_info=DEFAULT,
                _find_pkg_info=DEFAULT,
        ) as mocks:
            mocks['_find_git_info'].return_value = {
                'remotes': {},
                'tag': None,
                'commit': None,
                'dirty': None,
            }
            mocks['_find_pip_info'].side_effect = se_exception
            mocks['_find_pkg_info'].return_value = {
                'version': '1.2.3',
                'url': 'http://my.package.url/pkg_resources'
            }
            with patch('%s._git_repo_path' % pb,
                       new_callable=PropertyMock) as mock_is_git:
                mock_is_git.return_value = None
                res = self.cls.find_package_version()
        assert res.as_dict == VersionInfo(
            pkg_resources_version='1.2.3',
            pkg_resources_url='http://my.package.url/pkg_resources',
        ).as_dict
        assert mocks['_find_git_info'].mock_calls == []
        assert mocks['_find_pip_info'].mock_calls == [call(self.cls)]
        assert mocks['_find_pkg_info'].mock_calls == [call(self.cls)]
        assert mock_is_git.mock_calls == [call()]


class TestGitRepoPath(BaseTest):

    def test_true(self):
        with patch('%s.os.path.exists' % pbm) as mock_exists:
            with patch('%s._package_top_dir' % pb,
                       new_callable=PropertyMock) as mock_top_dir:
                mock_top_dir.return_value = ['/foo/bar', '/foo/bar/baz']
                mock_exists.side_effect = [False, True]
                res = self.cls._git_repo_path
        assert res == '/foo/bar/baz/.git'
        assert mock_exists.mock_calls == [
            call('/foo/bar/.git'),
            call('/foo/bar/baz/.git')
        ]

    def test_false(self):
        with patch('%s.os.path.exists' % pbm) as mock_exists:
            with patch('%s._package_top_dir' % pb,
                       new_callable=PropertyMock) as mock_top_dir:
                mock_top_dir.return_value = ['/foo/bar', '/foo/bar/baz']
                mock_exists.return_value = False
                res = self.cls._git_repo_path
        assert res is None
        assert mock_exists.mock_calls == [
            call('/foo/bar/.git'),
            call('/foo/bar/baz/.git')
        ]

    @pytest.mark.skip
    def test_not_yet(self):
        print("TODO: get rid of this naive package path finding, and"
              "do something better that actually looks for the package"
              "top directory")
        assert 1 == 0


class TestFindGitInfo(BaseTest):

    def test_find(self):
        with patch('%s.Repo' % pbm, autospec=True) as mock_repo:
            mock_repo.return_value = mockrepo(
                commit='12345678',
                dirty=False,
                remotes={
                    'origin': 'http://my.git/url',
                    'upstream': ['git@github.com:/foo/bar', 'bar'],
                    'foo': [],
                },
                tag='mytag'
            )
            res = self.cls._find_git_info('/git/repo/.git')
        assert res == {
            'commit': '12345678',
            'dirty': False,
            'tag': 'mytag',
            'remotes': {
                'origin': 'http://my.git/url',
                'upstream': 'git@github.com:/foo/bar'
            }
        }
        assert mock_repo.mock_calls == [
            call(path='/git/repo/.git', search_parent_directories=False),
            call().is_dirty(untracked_files=True)
        ]

    def test_no_git(self):

        def se_exc():
            raise Exception("foo")

        with patch('%s.Repo' % pbm, autospec=True) as mock_repo:
            mock_repo.side_effect = se_exc
            res = self.cls._find_git_info('/git/repo/.git')
        assert res == {
            'commit': None,
            'tag': None,
            'remotes': None,
            'dirty': None,
        }
        assert mock_repo.mock_calls == [
            call(path='/git/repo/.git', search_parent_directories=False)
        ]


class TestGetDistVersionUrl(BaseTest):

    def test_get(self):
        """pip 6.0+ - see issue #55"""
        dist = Mock(
            parsed_version=Version('1.2.3'),
            version='2.4.2',
            PKG_INFO='PKG-INFO',
        )
        metadata = [
            'Metadata-Version: 1.1',
            'Name: awslimitchecker',
            'Version: 0.1.0',
            'Summary: A script and python module to check your AWS service ',
            'Home-page: https://github.com/jantman/awslimitchecker',
            'Author: Jason Antman',
            'Author-email: jason@jasonantman.com',
            'License: UNKNOWN',
            'Description: awslimitchecker',
            '========================',
            '.. image:: https://pypip.in/v/awslimitchecker/badge.png',
            ':target: https://crate.io/packages/awslimitchecker',
            ':alt: PyPi package version',
            'Status',
            '------',
            'Keywords: AWS EC2 Amazon boto boto3 limits cloud',
            'Platform: UNKNOWN',
            'Classifier: Environment :: Console',
        ]

        def se_metadata(foo):
            for line in metadata:
                yield line

        dist.get_metadata_lines.side_effect = se_metadata
        res = self.cls._dist_version_url(dist)
        assert res == ('2.4.2',
                       'https://github.com/jantman/awslimitchecker')

    def test_pip1(self):
        """pip 1.x - see issue #54"""
        dist = Mock(
            parsed_version=('00000002', '00000004', '00000002', '*final'),
            version='2.4.2',
            PKG_INFO='PKG-INFO',
        )
        metadata = [
            'Metadata-Version: 1.1',
            'Name: awslimitchecker',
            'Version: 0.1.0',
            'Summary: A script and python module to check your AWS service ',
            'Home-page: https://github.com/jantman/awslimitchecker',
            'Author: Jason Antman',
            'Author-email: jason@jasonantman.com',
            'License: UNKNOWN',
            'Description: awslimitchecker',
            '========================',
            '.. image:: https://pypip.in/v/awslimitchecker/badge.png',
            ':target: https://crate.io/packages/awslimitchecker',
            ':alt: PyPi package version',
            'Status',
            '------',
            'Keywords: AWS EC2 Amazon boto boto3 limits cloud',
            'Platform: UNKNOWN',
            'Classifier: Environment :: Console',
        ]

        def se_metadata(foo):
            for line in metadata:
                yield line

        dist.get_metadata_lines.side_effect = se_metadata
        res = self.cls._dist_version_url(dist)
        assert res == ('2.4.2',
                       'https://github.com/jantman/awslimitchecker')

    def test_no_homepage(self):
        dist = Mock(
            version='1.2.3',
            PKG_INFO='PKG-INFO',
        )
        metadata = [
            'Metadata-Version: 1.1',
            'Name: awslimitchecker',
            'Version: 0.1.0',
            'Summary: A script and python module to check your AWS service ',
            'Author: Jason Antman',
            'Author-email: jason@jasonantman.com',
            'License: UNKNOWN',
            'Description: awslimitchecker',
            '========================',
            '.. image:: https://pypip.in/v/awslimitchecker/badge.png',
            ':target: https://crate.io/packages/awslimitchecker',
            ':alt: PyPi package version',
            'Status',
            '------',
            'Keywords: AWS EC2 Amazon boto boto3 limits cloud',
            'Platform: UNKNOWN',
            'Classifier: Environment :: Console',
        ]

        def se_metadata(foo):
            for line in metadata:
                yield line

        dist.get_metadata_lines.side_effect = se_metadata
        res = self.cls._dist_version_url(dist)
        assert res == ('1.2.3', None)


class TestFindPipInfo(BaseTest):

    def test_find(self):
        mock_distA = Mock(autospec=True, project_name='awslimitchecker')
        mock_distB = Mock(autospec=True, project_name='foo')
        mock_distC = Mock(autospec=True, project_name='another')
        installed_dists = [mock_distA, mock_distB, mock_distC]
        mock_frozen = Mock(
            autospec=True,
            req='foo==4.5.6'
        )

        with patch('%s.pip.get_installed_distributions' % pbm
                   ) as mock_pgid:
            with patch('%s.pip.FrozenRequirement.from_dist' % pbm
                       ) as mock_from_dist:
                with patch('%s._dist_version_url' % pb) as mock_dist_vu:
                    mock_pgid.return_value = installed_dists
                    mock_from_dist.return_value = mock_frozen
                    mock_dist_vu.return_value = ('4.5.6', 'http://foo')
                    res = self.cls._find_pip_info()
        assert res == {'version': '4.5.6', 'url': 'http://foo',
                       'requirement': 'foo==4.5.6'}
        assert mock_pgid.mock_calls == [call()]
        assert mock_from_dist.mock_calls == [call(mock_distB, [])]
        assert mock_dist_vu.mock_calls == [call(mock_distB)]

    def test_no_dist(self):
        mock_distB = Mock(autospec=True, project_name='other')
        mock_distC = Mock(autospec=True, project_name='another')
        installed_dists = [mock_distB, mock_distC]
        mock_frozen = Mock(
            autospec=True,
            req='awslimitchecker==0.1.0'
        )

        with patch('%s.pip.get_installed_distributions' % pbm
                   ) as mock_pgid:
            with patch('%s.pip.FrozenRequirement.from_dist' % pbm
                       ) as mock_from_dist:
                with patch('%s._dist_version_url' % pb) as mock_dist_vu:
                    mock_pgid.return_value = installed_dists
                    mock_from_dist.return_value = mock_frozen
                    mock_dist_vu.return_value = ('4.5.6', 'http://foo')
                    res = self.cls._find_pip_info()
        assert res == {}
        assert mock_pgid.mock_calls == [call()]
        assert mock_from_dist.mock_calls == []
        assert mock_dist_vu.mock_calls == []

    def test_req_https(self):
        req_str = 'git+https://github.com/jantman/foo.git@76c7e51' \
                  'f6e83350c72a1d3e8122ee03e589bbfde#egg=foo-master'
        mock_distA = Mock(autospec=True, project_name='awslimitchecker')
        mock_distB = Mock(autospec=True, project_name='foo')
        mock_distC = Mock(autospec=True, project_name='another')
        installed_dists = [mock_distA, mock_distB, mock_distC]
        mock_frozen = Mock(
            autospec=True,
            req=req_str
        )

        with patch('%s.pip.get_installed_distributions' % pbm
                   ) as mock_pgid:
            with patch('%s.pip.FrozenRequirement.from_dist' % pbm
                       ) as mock_from_dist:
                with patch('%s._dist_version_url' % pb) as mock_dist_vu:
                    mock_pgid.return_value = installed_dists
                    mock_from_dist.return_value = mock_frozen
                    mock_dist_vu.return_value = ('4.5.6', 'http://foo')
                    res = self.cls._find_pip_info()
        assert res == {'version': '4.5.6', 'url': 'http://foo',
                       'requirement': req_str}
        assert mock_pgid.mock_calls == [call()]
        assert mock_from_dist.mock_calls == [call(mock_distB, [])]
        assert mock_dist_vu.mock_calls == [call(mock_distB)]

    def test_req_git(self):
        req_str = 'git+git@github.com:jantman/foo.git@76c7e51f6e8' \
                  '3350c72a1d3e8122ee03e589bbfde#egg=foo-master'
        mock_distA = Mock(autospec=True, project_name='awslimitchecker')
        mock_distB = Mock(autospec=True, project_name='foo')
        mock_distC = Mock(autospec=True, project_name='another')
        installed_dists = [mock_distA, mock_distB, mock_distC]
        mock_frozen = Mock(
            autospec=True,
            req=req_str
        )

        with patch('%s.pip.get_installed_distributions' % pbm
                   ) as mock_pgid:
            with patch('%s.pip.FrozenRequirement.from_dist' % pbm
                       ) as mock_from_dist:
                with patch('%s._dist_version_url' % pb) as mock_dist_vu:
                    mock_pgid.return_value = installed_dists
                    mock_from_dist.return_value = mock_frozen
                    mock_dist_vu.return_value = ('4.5.6', 'http://foo')
                    res = self.cls._find_pip_info()
        assert res == {'version': '4.5.6', 'url': 'http://foo',
                       'requirement': req_str}
        assert mock_pgid.mock_calls == [call()]
        assert mock_from_dist.mock_calls == [call(mock_distB, [])]
        assert mock_dist_vu.mock_calls == [call(mock_distB)]


class TestFindPkgInfo(BaseTest):

    def test_find_pkg_info(self):
        mock_distA = Mock(autospec=True, project_name='awslimitchecker')
        with patch('%s.pkg_resources.require' % pbm) as mock_require:
            with patch('%s._dist_version_url' % pb) as mock_dvu:
                mock_require.return_value = [mock_distA]
                mock_dvu.return_value = ('7.8.9', 'http://foobar')
                res = self.cls._find_pkg_info()
        assert res == {'version': '7.8.9', 'url': 'http://foobar'}


class TestPackageTopDir(BaseTest):

    def test_none(self):
        self.cls.package_dir = '/foo'
        self.cls._pip_locations = []
        self.cls._pkg_resources_locations = []
        assert self.cls._package_top_dir == ['/foo']

    def test_pip(self):
        self.cls.package_dir = '/foo'
        self.cls._pip_locations = ['/bar', '/baz', None]
        self.cls._pkg_resources_locations = []
        assert self.cls._package_top_dir == ['/bar', '/baz', '/foo']

    def test_pkg_resources(self):
        self.cls.package_dir = '/foo'
        self.cls._pip_locations = []
        self.cls._pkg_resources_locations = ['/bar', '/baz', None]
        assert self.cls._package_top_dir == ['/bar', '/baz', '/foo']

    def test_all(self):
        self.cls.package_dir = '/foo'
        self.cls._pip_locations = ['/bar', None]
        self.cls._pkg_resources_locations = ['/baz', None]
        assert self.cls._package_top_dir == ['/bar', '/baz', '/foo']


class TestChdir(object):

    def test_chdir(self):
        with patch('%s.os.getcwd' % pbm, autospec=True) as mock_getcwd:
            with patch('%s.os.chdir' % pbm, autospec=True) as mock_chdir:
                mock_getcwd.return_value = '/old/cwd'
                with chdir('/new/dir'):
                    pass
        assert mock_chdir.mock_calls == [
            call('/new/dir'),
            call('/old/cwd')
        ]



"""
versionfinder/tests/test_acceptance.py

SEE TestAcceptance class docstring for information!

!!!!!!!IMPORTANT!!!!!!
When adding test cases, also add them to the list in README.rst!

The latest version of this package is available at:
<https://github.com/jantman/versionfinder>

################################################################################
Copyright 2015 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of versionfinder.

    versionfinder is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    versionfinder is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with versionfinder.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the GPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/versionfinder> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
import sys
import os
import subprocess
import json
import inspect
import backoff
import requests
import logging
from versionfinder.versionfinder import chdir
from tempfile import mkdtemp

if sys.version_info >= (3, 3):
    from subprocess import DEVNULL
else:  # unreachable under 3.4+ - pragma: no cover
    DEVNULL = open(os.devnull, 'wb')

logger = logging.getLogger(__name__)

TEST_PROJECT = 'versionfinder_test_pkg'
TEST_GIT_HTTPS_URL = 'https://github.com/jantman/versionfinder-test-pkg.git'
TEST_FORK_HTTPS_URL = 'https://github.com/sniknej/versionfinder-test-pkg.git'
TEST_PROJECT_URL = 'https://github.com/jantman/versionfinder-test-pkg'
TEST_VERSION = '0.2.5'
TEST_TAG = '0.2.5'
TEST_TAG_COMMIT = '212d9950ca546013b5539133d7862c2fa4bea6f3'
TEST_MASTER_COMMIT = '033d10a4280311008b85a4af4cc845e0d11b76b0'
TEST_BRANCH = 'testbranch'
TEST_BRANCH_COMMIT = 'a5f7fa6b3b7aa57aa2661633896ba61eb78bbc97'

TEST_TARBALL = 'https://github.com/jantman/versionfinder-test-pkg/releases/' \
               'download/{ver}/versionfinder_test_pkg-{ver}.tar' \
               '.gz'.format(ver=TEST_VERSION)
TEST_TARBALL_PATH = None  # set by setup_module()
TEST_WHEEL = 'https://github.com/jantman/versionfinder-test-pkg/releases/' \
             'download/{ver}/versionfinder_test_pkg-{ver}-py2.py3-none-any' \
             '.whl'.format(ver=TEST_VERSION)
TEST_WHEEL_PATH = None  # set by setup_module()
TEST_EGG_FMT = 'https://github.com/jantman/versionfinder-test-pkg/releases/' \
               'download/{ver}/versionfinder_test_pkg-{ver}-%s.egg'.format(
                   ver=TEST_VERSION)
TEST_EGG_VERSIONS = ['py2.6', 'py2.7', 'py3.3', 'py3.4', 'py3.5']
TEST_EGG_PATH = None


def setup_module(module):
    global TEST_TARBALL_PATH, TEST_WHEEL_PATH, TEST_EGG_PATH
    TEST_TARBALL_PATH = get_package(TEST_TARBALL)
    TEST_WHEEL_PATH = get_package(TEST_WHEEL)
    pyver = 'py%s.%s' % (sys.version_info[0], sys.version_info[1])
    if pyver in TEST_EGG_VERSIONS:
        TEST_EGG_PATH = get_package(TEST_EGG_FMT % pyver)
    else:
        print("'%s' not in TEST_EGG_VERSIONS" % pyver)


def _get_git_commit():
    """
    Get the current (short) git commit hash of the current directory.

    :returns: short git hash
    :rtype: string
    """
    try:
        commit = _check_output([
            'git',
            'rev-parse',
            'HEAD'
        ], stderr=DEVNULL).strip()
        logger.debug("Found source git commit: %s", commit)
    except Exception:
        logger.debug("Unable to run git to get commit")
        commit = None
    return commit


def _check_output(args, stderr=None, env=None):
    """
    Python version compatibility wrapper for subprocess.check_output

    :param stderr: what to do with STDERR - None or an appropriate argument
     to subprocess.check_output / subprocess.Popen
    :raises: subprocess.CalledProcessError
    :returns: command output
    :rtype: string
    """
    p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=stderr, env=env)
    (res, err) = p.communicate()
    if p.returncode != 0:
        print("Command %s OUT/ERR:\n%s", args, res)
        raise subprocess.CalledProcessError(p.returncode, args)
    if not isinstance(res, type("")):
        res = res.decode('utf-8')
    return res


class AcceptanceHelpers(object):
    """
    Long-running acceptance tests for VersionFinder.

    The purpose of these tests is to create virtualenvs and then install an
    example package (https://github.com/jantman/versionfinder-test-pkg) in them
    using a variety of methods (i.e. setup.py, pip install from directory,
    git URL, tarball or wheel, etc.), and then call the package's console entry
    point (``versionfinder-test``) which calls ``versionfinder.find_version()``
    from a number of different locations in the package tree, and finally writes
    JSON to STDOUT describing the result of each of the ``find_version()`` calls
    (and any exceptions they raised). We then compare them to each other and,
    assuming they're all identical, compare them to our expected output.

    The goal is to confirm that versionfinder returns the correct information
    when the target package is installed with as many different methods as
    possible.

    This class defines a bunch of helper methods, which are used in the tests
    (in subclasses).
    """

    def setup_method(self, method):
        os.environ['VERSIONCHECK_DEBUG'] = 'true'
        print("\n")
        self.source_dir = self._get_source_dir()
        self._set_git_config()
        self.current_venv_path = sys.prefix

    def _get_source_dir(self):
        """
        Determine the directory containing the project source. This is assumed
        to be either the TOXINIDIR environment variable, or determined relative
        to this file.

        :returns: path to the awslimitchecker source
        :rtype: str
        """

        s = os.environ.get('TOXINIDIR', None)
        if s is None:
            s = os.path.abspath(
                os.path.join(
                    os.path.dirname(os.path.abspath(__file__)),
                    '..',
                    '..'
                )
            )
        assert os.path.exists(s)
        return s

    def _git_add_remote(self, path, rmt_name, rmt_url):
        """
        Add a git remote to the repo at path

        :param path: path to the git repo
        :type path: str
        :param rmt_name: name for the remote
        :type rmt_name: str
        :param rmt_url: URL for the remote
        :type rmt_url: str
        """
        args = [
            'git',
            'remote',
            'add',
            rmt_name,
            rmt_url
        ]
        with chdir(path):
            res = _check_output(args)
            print(res)

    def _set_git_config(self, set_in_travis=False):
        if not set_in_travis and os.environ.get('TRAVIS', '') != 'true':
            print("not running in Travis; not setting git config")
            return
        try:
            res = _check_output([
                'git',
                'config',
                'user.email'
            ]).strip()
        except subprocess.CalledProcessError as ex:
            res = None
        if res != '' and res is not None:
            print("Got git config user.email as %s" % res)
        else:
            res = _check_output([
                'git',
                'config',
                'user.email',
                'travisci@jasonantman.com'
            ])
            print("Set git config user.email:\n%s" % res)
        # name
        try:
            res = _check_output([
                'git',
                'config',
                'user.name'
            ]).strip()
        except subprocess.CalledProcessError as ex:
            print(ex)
            res = None
        if res != '' and res is not None:
            print("Got git config user.name as %s" % res)
        else:
            res = _check_output([
                'git',
                'config',
                'user.name',
                'travisci'
            ])
            print("Set git config user.name:\n%s" % res)

    def _set_git_tag(self, path, tagname):
        """set a git tag for the current commit"""
        with chdir(path):
            self._set_git_config()
            commit = _check_output([
                'git',
                'rev-parse',
                'HEAD'
            ]).strip()
            print("Creating git tag 'versiontest' of %s" % commit)
            res = _check_output([
                'git',
                'tag',
                '-a',
                '-m',
                tagname,
                tagname
            ])
            print(res)
            print("Source git tag: %s" % tagname)
        return tagname

    def _git_add_commit(self, path, commit_msg):
        """
        git add -A and git commit -m commit_msg in ``path``.
        """
        print_header('_git_add_commit(%s, %s)' % (path, commit_msg))
        with chdir(path):
            self._set_git_config()
            out = _check_output([
                'git',
                'add',
                '-A'
            ])
            print(out)
            out = _check_output([
                'git',
                'commit',
                '-m',
                '"%s"' % commit_msg
            ])
            print(out)
            out = _check_output([
                'git',
                'rev-parse',
                'HEAD'
            ]).strip()
        return out

    def _make_venv(self, path):
        """
        Create a venv in ``path``. Make sure it exists.

        :param path: filesystem path to directory to make the venv base
        """
        virtualenv = os.path.join(self.current_venv_path, 'bin', 'virtualenv')
        assert os.path.exists(virtualenv) is True, 'virtualenv not found'
        args = [virtualenv, path]
        print_header(" _make_venv() running: " + ' '.join(args))
        try:
            res = _check_output(args, stderr=subprocess.STDOUT)
            print(res)
            print_header("DONE")
        except subprocess.CalledProcessError:
            print_header('FAILED')
        pypath = os.path.join(path, 'bin', 'python')
        assert os.path.exists(pypath) is True, "does not exist: %s" % pypath
        # install our source in the venv
        self._pip_install(path, [self.source_dir])

    def _pip_install(self, path, args):
        """
        In the virtualenv at ``path``, run ``pip install [args]``.

        :param path: venv base/root path
        :type path: str
        :param args: ``pip install`` arguments
        :type args: list
        """
        pip = os.path.join(path, 'bin', 'pip')
        # get pip version
        res = _check_output([pip, '--version']).strip()
        final_args = [pip, 'install']
        final_args.extend(args)
        print_header("_pip_install() running: " + ' '.join(final_args))
        res = _check_output(final_args, stderr=subprocess.STDOUT)
        print(res)
        print_header('DONE')

    def _easy_install(self, path, args):
        """
        In the virtualenv at ``path``, run ``easy_install [args]``.

        :param path: venv base/root path
        :type path: str
        :param args: ``easy_install`` arguments
        :type args: list
        """
        easy_install = os.path.join(path, 'bin', 'easy_install')
        final_args = [easy_install]
        final_args.extend(args)
        print_header("_easy_install() running: " + ' '.join(final_args))
        res = _check_output(final_args, stderr=subprocess.STDOUT)
        print(res)
        print_header('DONE')

    def _make_git_repo(self, path):
        """create a git repo under path; return the commit"""
        print_header("creating git repository in %s" % path)
        with chdir(path):
            _check_output(['git', 'init', '.'])
            self._set_git_config()
            with open('foo', 'w') as fh:
                fh.write('foo')
            _check_output(['git', 'add', 'foo'])
            self._set_git_config(set_in_travis=True)
            _check_output(['git', 'commit', '-m', 'foo'])
            commit = _get_git_commit()
            print_header("git repository in %s commit: %s" % (path, commit))
        return commit

    def _get_version(self, path, env=None):
        """
        In the virtualenv at ``path``, run ``versionfinder-test`` and
        return the JSON-decoded output dict.

        :param path: venv base/root path
        :type path: str
        :return: versionfinder-test command output
        :rtype: dict
        """
        args = [os.path.join(path, 'bin', 'versionfinder-test')]
        print_header("_get_version() running: " + ' '.join(args))
        res = _check_output(args, stderr=subprocess.STDOUT, env=env)
        print(res)
        print('DONE')
        j = json.loads(res.strip().split("\n")[-1])
        return strip_unicode(j)

    def _get_result(self, d):
        """
        Given the raw (JSON-decoded) result dict from :py:meth:`~._get_version`,
        iterate through all of the keys. Assert that their values are all
        identical, and don't contain exceptions. Return the identical result
        dict.

        :param d: result dict from :py:meth:`~._get_version`
        :type d: dict
        :return: dict describing identical results
        :rtype: dict
        """
        keys = sorted(d.keys())
        # use the first one as "expected" to compare the others against
        expected = d[keys[0]]
        err = ''
        # iterate each result; it's an error if either it failed, or it doesn't
        # match the expected value
        for k in keys:
            if d[k].get('failed', True) is True:
                err += "Key %s failed with %s %s: \n%s\n" % (
                    k, d[k].get('exc_type', ''), d[k].get('exc_str', ''),
                    d[k].get('traceback', ''))
            if d[k] != expected:
                err += "Key %s does not match expected:\n%s\n" % (
                    k, dictdiff(d[k], expected))
        # AssertionError on any error conditions, but we want to show ALL
        assert err == '', err
        # else return the indentical dict for all of them
        return expected

    def _git_checkout(self, path, ref):
        cmd = ['git', 'checkout', '-f', ref]
        print_header("_git_checkout() running: '%s' in: %s" % (
            ' '.join(cmd), path))
        with chdir(path):
            output = _check_output(cmd, stderr=subprocess.STDOUT)
            print(output)
        print_header('DONE')

    def _git_clone_test(self, ref=None, url=None):
        """
        Clone TEST_GIT_HTTPS_URL to a local temporary directory; checkout
        the specified ref.

        :return: path to git clone
        :rtype: str
        """
        if url is None:
            url = TEST_GIT_HTTPS_URL
        d = mkdtemp(prefix='pytest-versionfinder')
        print_header('_git_clone_test(%s) cloning %s into %s' % (
            ref, url, d))
        output = _check_output(
            ['git', 'clone', url, d],
            stderr=subprocess.STDOUT
        )
        print(output)
        print_header('DONE')
        if ref is not None:
            self._git_checkout(d, ref)
        return d


@pytest.mark.acceptance
class TestPip(AcceptanceHelpers):

    def test_install_local_master(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        test_src = self._git_clone_test()
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, test_src))
        self._pip_install(path, [test_src])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': None,
                'git_tag': None,
                'git_remotes': None,
                'git_is_dirty': None,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'versionfinder-test-pkg==%s' % TEST_VERSION,
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_local_e(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        test_src = self._git_clone_test()
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, test_src))
        self._pip_install(path, ['-e', test_src])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': TEST_MASTER_COMMIT,
                'git_tag': None,
                'git_remotes': {
                    'origin': TEST_GIT_HTTPS_URL,
                },
                'git_is_dirty': False,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
                'pip_requirement': 'git+%s@%s#egg=%s' % (
                    TEST_GIT_HTTPS_URL, TEST_MASTER_COMMIT, TEST_PROJECT
                )
            }
        }
        assert actual == expected

    def test_install_local_e_dirty(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        test_src = self._git_clone_test()
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, test_src))
        self._pip_install(path, ['-e', test_src])
        fpath = os.path.join(test_src, 'versionfinder_test_pkg', 'foo.py')
        print("Creating junk file at %s" % fpath)
        with open(fpath, 'w') as fh:
            fh.write("testing")
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': TEST_MASTER_COMMIT,
                'git_tag': None,
                'git_remotes': {
                    'origin': TEST_GIT_HTTPS_URL,
                },
                'git_is_dirty': True,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'git+%s@%s#egg=%s' % (
                    TEST_GIT_HTTPS_URL, TEST_MASTER_COMMIT, TEST_PROJECT
                ),
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_local_e_tag(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        test_src = self._git_clone_test(ref=TEST_TAG_COMMIT)
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, test_src))
        self._pip_install(path, ['-e', test_src])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': TEST_TAG_COMMIT,
                'git_tag': TEST_TAG,
                'git_remotes': {
                    'origin': TEST_GIT_HTTPS_URL,
                },
                'git_is_dirty': False,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'git+%s@%s#egg=versionfinder_test_pkg' % (
                    TEST_GIT_HTTPS_URL, TEST_TAG_COMMIT
                ),
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_local_e_checkout_tag(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        test_src = self._git_clone_test()
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, test_src))
        self._pip_install(path, ['-e', test_src])
        self._git_checkout(test_src, TEST_TAG)
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': TEST_TAG_COMMIT,
                'git_tag': TEST_TAG,
                'git_remotes': {
                    'origin': TEST_GIT_HTTPS_URL,
                },
                'git_is_dirty': False,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'git+%s@%s#egg=versionfinder_test_pkg' % (
                    TEST_GIT_HTTPS_URL, TEST_TAG_COMMIT
                ),
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_local_e_checkout_commit(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        test_src = self._git_clone_test()
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, test_src))
        self._pip_install(path, ['-e', test_src])
        self._git_checkout(test_src, TEST_TAG_COMMIT)
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': TEST_TAG_COMMIT,
                'git_tag': TEST_TAG,
                'git_remotes': {
                    'origin': TEST_GIT_HTTPS_URL,
                },
                'git_is_dirty': False,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'git+%s@%s#egg=versionfinder_test_pkg' % (
                    TEST_GIT_HTTPS_URL, TEST_TAG_COMMIT
                ),
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_local_e_multiple_remotes(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        test_src = self._git_clone_test()
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, test_src))
        self._git_add_remote(test_src, 'testremote',
                             'https://github.com/jantman/awslimitchecker.git')
        self._pip_install(path, ['-e', test_src])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': TEST_MASTER_COMMIT,
                'git_tag': None,
                'git_remotes': {
                    'origin': TEST_GIT_HTTPS_URL,
                    'testremote': 'https://github.com/jantman/'
                                  'awslimitchecker.git',
                },
                'git_is_dirty': False,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'git+%s@%s#egg=versionfinder_test_pkg' % (
                    TEST_GIT_HTTPS_URL, TEST_MASTER_COMMIT
                ),
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_sdist(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, TEST_TARBALL_PATH))
        self._pip_install(path, [TEST_TARBALL_PATH])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': None,
                'git_tag': None,
                'git_remotes': None,
                'git_is_dirty': None,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'versionfinder-test-pkg==%s' % TEST_VERSION,
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    @pytest.mark.skipif(
        sys.version_info[0:2] >= (3, 6),
        reason='pip 1.5.4 does not work on py >= 3.6'
    )
    def test_install_sdist_pip154(self, capsys, tmpdir):
        """regression test for issue #55"""
        path = str(tmpdir)
        self._make_venv(path)
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, TEST_TARBALL_PATH))
        self._pip_install(path, ['--force-reinstall', 'pip==1.5.4'])
        self._pip_install(path, [TEST_TARBALL_PATH])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': None,
                'git_tag': None,
                'git_remotes': None,
                'git_is_dirty': None,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'versionfinder-test-pkg==%s' % TEST_VERSION,
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_bdist_wheel(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, TEST_WHEEL_PATH))
        self._pip_install(path, [TEST_WHEEL_PATH])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': None,
                'git_tag': None,
                'git_remotes': None,
                'git_is_dirty': None,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'versionfinder-test-pkg==%s' % TEST_VERSION,
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_bdist_wheel_no_git_binary(self, capsys, tmpdir):
        e = {x: os.environ[x] for x in os.environ}
        e['GIT_PYTHON_GIT_EXECUTABLE'] = '/tmp/NoSuchFile'
        path = str(tmpdir)
        self._make_venv(path)
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, TEST_WHEEL_PATH))
        self._pip_install(path, [TEST_WHEEL_PATH])
        actual = self._get_result(self._get_version(path, env=e))
        expected = {
            'failed': False,
            'result': {
                'git_commit': None,
                'git_tag': None,
                'git_remotes': None,
                'git_is_dirty': None,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'versionfinder-test-pkg==%s' % TEST_VERSION,
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_git(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, 'git'))
        self._pip_install(path, [
            'git+%s#egg=versionfinder-test-pkg' % (
                TEST_GIT_HTTPS_URL
            )
        ])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': None,
                'git_tag': None,
                'git_remotes': None,
                'git_is_dirty': None,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'versionfinder-test-pkg==%s' % TEST_VERSION,
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_git_fork(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, 'git'))
        self._pip_install(path, [
            'git+%s#egg=versionfinder-test-pkg' % (
                TEST_FORK_HTTPS_URL
            )
        ])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': None,
                'git_tag': None,
                'git_remotes': None,
                'git_is_dirty': None,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'versionfinder-test-pkg==%s' % TEST_VERSION,
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_git_commit(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, 'git'))
        self._pip_install(path, [
            'git+%s@%s#egg=versionfinder-test-pkg' % (
                TEST_GIT_HTTPS_URL,
                TEST_MASTER_COMMIT
            )
        ])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': None,
                'git_tag': None,
                'git_remotes': None,
                'git_is_dirty': None,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'versionfinder-test-pkg==%s' % TEST_VERSION,
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_git_tag(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, 'git'))
        self._pip_install(path, [
            'git+%s@%s#egg=versionfinder-test-pkg' % (
                TEST_GIT_HTTPS_URL,
                TEST_TAG
            )
        ])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': None,
                'git_tag': None,
                'git_remotes': None,
                'git_is_dirty': None,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'versionfinder-test-pkg==%s' % TEST_VERSION,
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_git_branch(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, 'git'))
        self._pip_install(path, [
            'git+%s@%s#egg=versionfinder-test-pkg' % (
                TEST_GIT_HTTPS_URL,
                TEST_BRANCH
            )
        ])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': None,
                'git_tag': None,
                'git_remotes': None,
                'git_is_dirty': None,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'versionfinder-test-pkg==%s' % TEST_VERSION,
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_git_e(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, 'git'))
        self._pip_install(path, [
            '-e',
            'git+%s#egg=versionfinder-test-pkg' % (
                TEST_GIT_HTTPS_URL
            )
        ])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': TEST_MASTER_COMMIT,
                'git_tag': None,
                'git_remotes': {
                    'origin': TEST_GIT_HTTPS_URL,
                },
                'git_is_dirty': False,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'git+%s@%s#egg=versionfinder_test_pkg' % (
                    TEST_GIT_HTTPS_URL, TEST_MASTER_COMMIT
                ),
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_git_e_fork(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, 'git'))
        self._pip_install(path, [
            '-e',
            'git+%s#egg=versionfinder-test-pkg' % (
                TEST_FORK_HTTPS_URL
            )
        ])
        self._git_add_remote(
            os.path.join(path, 'src', 'versionfinder-test-pkg'),
            'upstream',
            TEST_GIT_HTTPS_URL
        )
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': TEST_MASTER_COMMIT,
                'git_tag': None,
                'git_remotes': {
                    'origin': TEST_FORK_HTTPS_URL,
                    'upstream': TEST_GIT_HTTPS_URL,
                },
                'git_is_dirty': False,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'git+%s@%s#egg=versionfinder_test_pkg' % (
                    TEST_FORK_HTTPS_URL, TEST_MASTER_COMMIT
                ),
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_git_e_multiple_remotes(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, 'git'))
        self._pip_install(path, [
            '-e',
            'git+%s#egg=versionfinder-test-pkg' % (
                TEST_GIT_HTTPS_URL
            )
        ])
        p = os.path.join(path, 'src', 'versionfinder-test-pkg')
        self._git_add_remote(p, 'testremote',
                             TEST_FORK_HTTPS_URL)
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': TEST_MASTER_COMMIT,
                'git_tag': None,
                'git_remotes': {
                    'origin': TEST_GIT_HTTPS_URL,
                    'testremote': TEST_FORK_HTTPS_URL
                },
                'git_is_dirty': False,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'git+%s@%s#egg=versionfinder_test_pkg' % (
                    TEST_GIT_HTTPS_URL, TEST_MASTER_COMMIT
                ),
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_git_e_dirty(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, 'git'))
        self._pip_install(path, [
            '-e',
            'git+%s#egg=versionfinder-test-pkg' % (
                TEST_GIT_HTTPS_URL
            )
        ])
        fpath = os.path.join(
            path, 'src', 'versionfinder-test-pkg', 'versionfinder_test_pkg',
            'foo.py'
        )
        print("Creating junk file at %s" % fpath)
        with open(fpath, 'w') as fh:
            fh.write("testing")
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': TEST_MASTER_COMMIT,
                'git_tag': None,
                'git_remotes': {
                    'origin': TEST_GIT_HTTPS_URL,
                },
                'git_is_dirty': True,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'git+%s@%s#egg=versionfinder_test_pkg' % (
                    TEST_GIT_HTTPS_URL, TEST_MASTER_COMMIT
                ),
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_git_e_commit(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, 'git'))
        self._pip_install(path, [
            '-e',
            'git+%s@%s#egg=versionfinder-test-pkg' % (
                TEST_GIT_HTTPS_URL,
                TEST_MASTER_COMMIT
            )
        ])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': TEST_MASTER_COMMIT,
                'git_tag': None,
                'git_remotes': {
                    'origin': TEST_GIT_HTTPS_URL,
                },
                'git_is_dirty': False,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'git+%s@%s#egg=versionfinder_test_pkg' % (
                    TEST_GIT_HTTPS_URL, TEST_MASTER_COMMIT
                ),
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_git_e_tag(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, 'git'))
        self._pip_install(path, [
            '-e',
            'git+%s@%s#egg=versionfinder-test-pkg' % (
                TEST_GIT_HTTPS_URL,
                TEST_TAG
            )
        ])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': TEST_TAG_COMMIT,
                'git_tag': TEST_TAG,
                'git_remotes': {
                    'origin': TEST_GIT_HTTPS_URL,
                },
                'git_is_dirty': False,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'git+%s@%s#egg=versionfinder_test_pkg' % (
                    TEST_GIT_HTTPS_URL, TEST_TAG_COMMIT
                ),
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_git_e_branch(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, 'git'))
        self._pip_install(path, [
            '-e',
            'git+%s@%s#egg=versionfinder-test-pkg' % (
                TEST_GIT_HTTPS_URL,
                TEST_BRANCH
            )
        ])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': TEST_BRANCH_COMMIT,
                'git_tag': None,
                'git_remotes': {
                    'origin': TEST_GIT_HTTPS_URL,
                },
                'git_is_dirty': False,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'git+%s@%s#egg=versionfinder_test_pkg' % (
                    TEST_GIT_HTTPS_URL, TEST_BRANCH_COMMIT
                ),
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_sdist_in_git_repo(self, capsys, tmpdir):
        """regression test for issue #73"""
        path = str(tmpdir)
        self._make_git_repo(path)
        self._make_venv(path)
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, TEST_TARBALL_PATH))
        self._pip_install(path, [TEST_TARBALL_PATH])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': None,
                'git_tag': None,
                'git_remotes': None,
                'git_is_dirty': None,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'versionfinder-test-pkg==%s' % TEST_VERSION,
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_wheel_in_git_repo(self, capsys, tmpdir):
        """regression test for issue #73"""
        path = str(tmpdir)
        self._make_git_repo(path)
        self._make_venv(path)
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, TEST_WHEEL_PATH))
        self._pip_install(path, [TEST_WHEEL_PATH])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': None,
                'git_tag': None,
                'git_remotes': None,
                'git_is_dirty': None,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'versionfinder-test-pkg==%s' % TEST_VERSION,
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected


@pytest.mark.acceptance
class TestSetupPy(AcceptanceHelpers):

    def test_develop(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        test_src = self._git_clone_test()
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, test_src))
        with chdir(test_src):
            cmd = [
                os.path.join(path, 'bin', 'python'),
                os.path.join(test_src, 'setup.py'),
                'develop'
            ]
            print_header('running: %s' % ' '.join(cmd))
            output = _check_output(cmd, stderr=subprocess.STDOUT)
            print(output)
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': TEST_MASTER_COMMIT,
                'git_tag': None,
                'git_remotes': {
                    'origin': TEST_GIT_HTTPS_URL,
                },
                'git_is_dirty': False,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'git+%s@%s#egg=versionfinder_test_pkg' % (
                    TEST_GIT_HTTPS_URL, TEST_MASTER_COMMIT
                ),
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        test_src = self._git_clone_test()
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, test_src))
        with chdir(test_src):
            cmd = [
                os.path.join(path, 'bin', 'python'),
                os.path.join(test_src, 'setup.py'),
                'install'
            ]
            print_header('running: %s' % ' '.join(cmd))
            output = _check_output(cmd, stderr=subprocess.STDOUT)
            print(output)
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': None,
                'git_tag': None,
                'git_remotes': None,
                'git_is_dirty': None,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'versionfinder-test-pkg==%s' % TEST_VERSION,
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected


@pytest.mark.acceptance
class TestEasyInstall(AcceptanceHelpers):

    def test_install_local_master(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        test_src = self._git_clone_test()
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, test_src))
        self._easy_install(path, [test_src])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': None,
                'git_tag': None,
                'git_remotes': None,
                'git_is_dirty': None,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'versionfinder-test-pkg==%s' % TEST_VERSION,
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_sdist(self, capsys, tmpdir):
        """regression test for issue #73"""
        path = str(tmpdir)
        self._make_venv(path)
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, TEST_TARBALL_PATH))
        self._easy_install(path, [TEST_TARBALL_PATH])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': None,
                'git_tag': None,
                'git_remotes': None,
                'git_is_dirty': None,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'versionfinder-test-pkg==%s' % TEST_VERSION,
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_egg(self, capsys, tmpdir):
        """regression test for issue #73"""
        if TEST_EGG_PATH is None:
            pytest.skip("No egg for python version")
        path = str(tmpdir)
        self._make_venv(path)
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, TEST_EGG_PATH))
        self._easy_install(path, [TEST_TARBALL_PATH])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': None,
                'git_tag': None,
                'git_remotes': None,
                'git_is_dirty': None,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'versionfinder-test-pkg==%s' % TEST_VERSION,
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_local_master_in_git_repo(self, capsys, tmpdir):
        path = str(tmpdir)
        self._make_venv(path)
        self._make_git_repo(path)
        test_src = self._git_clone_test()
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, test_src))
        self._easy_install(path, [test_src])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': None,
                'git_tag': None,
                'git_remotes': None,
                'git_is_dirty': None,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'versionfinder-test-pkg==%s' % TEST_VERSION,
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_sdist_in_git_repo(self, capsys, tmpdir):
        """regression test for issue #73"""
        path = str(tmpdir)
        self._make_venv(path)
        self._make_git_repo(path)
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, TEST_TARBALL_PATH))
        self._easy_install(path, [TEST_TARBALL_PATH])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': None,
                'git_tag': None,
                'git_remotes': None,
                'git_is_dirty': None,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'versionfinder-test-pkg==%s' % TEST_VERSION,
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected

    def test_install_egg_in_git_repo(self, capsys, tmpdir):
        """regression test for issue #73"""
        if TEST_EGG_PATH is None:
            pytest.skip("No egg for python version")
        path = str(tmpdir)
        self._make_venv(path)
        self._make_git_repo(path)
        with capsys.disabled():
            print("\n%s() venv=%s src=%s" % (
                inspect.stack()[0][0].f_code.co_name, path, TEST_EGG_PATH))
        self._easy_install(path, [TEST_TARBALL_PATH])
        actual = self._get_result(self._get_version(path))
        expected = {
            'failed': False,
            'result': {
                'git_commit': None,
                'git_tag': None,
                'git_remotes': None,
                'git_is_dirty': None,
                'pip_version': TEST_VERSION,
                'pip_url': TEST_PROJECT_URL,
                'pip_requirement': 'versionfinder-test-pkg==%s' % TEST_VERSION,
                'pkg_resources_version': TEST_VERSION,
                'pkg_resources_url': TEST_PROJECT_URL,
            }
        }
        assert actual == expected


def get_package(pkg_url):
    """
    Download the package from ``pkg_url`` to a tempdir.

    :param pkg_url: url of the package to download
    :type pkg_url: str
    :return: path to the package on disk
    :rtype: str
    """
    fname = pkg_url.split('/')[-1]
    d = mkdtemp(prefix='pytest-versionfinder')
    p = os.path.join(d, fname)
    r = requests_get(pkg_url)
    assert r.status_code == 200
    with open(p, 'wb') as fh:
        for chunk in r:
            fh.write(chunk)
    return p


# try up to 10 times to get a response that isn't rate-limited
@backoff.on_predicate(backoff.expo,
                      lambda r: r.status_code == 429, max_tries=10)
def requests_get(pkg_url):
    r = requests.get(pkg_url, stream=True)
    return r


def dictdiff(actual, expected, prefix=None):
    s = ''
    keys = set(actual.keys() + expected.keys())
    for k in sorted(keys):
        a = actual.get(k, '<missing>')
        e = expected.get(k, '<missing>')
        if prefix is None:
            k_str = k
        else:
            k_str = prefix + '->' + k
        if isinstance(a, type({})) and isinstance(e, type({})):
            s += dictdiff(a, e, prefix=k)
        else:
            if a != e:
                s += "*%s: actual='%s' expected='%s'\n" % (k_str, a, e)
            else:
                s += "%s: actual='%s' expected='%s'\n" % (k_str, a, e)
    return s


def print_header(s):
    print("%s %s %s" % ("#" * 20, s, "#" * 20))


def strip_unicode(d):
    n = {}
    for k, v in d.items():
        if isinstance(v, type({})):
            n[str(k)] = strip_unicode(v)
        elif v is True or v is False or v is None:
            n[str(k)] = v
        else:
            n[str(k)] = str(v)
    return n



"""
The latest version of this package is available at:
<http://github.com/jantman/versionfinder>

##################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of versionfinder.

    versionfinder is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    versionfinder is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with versionfinder.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
##################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/versionfinder> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
##################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
##################################################################################
"""

import inspect
from .versionfinder import VersionFinder


def find_version(*args, **kwargs):
    """
    Wrapper around :py:class:`~.VersionFinder` and its
    :py:meth:`~.VersionFinder.find_package_version` method. Pass arguments and
    kwargs to VersionFinder constructor, return the value of its
    ``find_package_version`` method.

    :param package_name: name of the package to find information about
    :type package_name: str
    :param package_file: absolute path to a Python source file in the
      package to find information about; if not specified, the file calling
      this class will be used
    :type package_file: str
    :param log: If not set to True, the "versionfinder" and "pip" loggers
      will be set to a level of ``logging.CRITICAL`` to suppress
      log output. If set to True, you will see a LOT of debug-level log
      output, for debugging the internals of versionfinder.
    :type log: bool
    :returns: information about the installed version of the package
    :rtype: :py:class:`~versionfinder.versioninfo.VersionInfo`
    """
    if 'caller_frame' not in kwargs:
        kwargs['caller_frame'] = inspect.stack()[1][0]
    return VersionFinder(*args, **kwargs).find_package_version()



"""
versionfinder/versioninfo.py

The latest version of this package is available at:
<https://github.com/jantman/versionfinder>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of versionfinder.

    versionfinder is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    versionfinder is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with versionfinder.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the GPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/versionfinder> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""


class VersionInfo(object):
    """
    Class describing :py:class:`~.VersionFinder` result; the discovered
    information about the version and source of an installed package.
    """

    def __init__(self, pip_version=None, pip_url=None, pip_requirement=None,
                 pkg_resources_version=None, pkg_resources_url=None,
                 git_tag=None, git_commit=None, git_remotes=None,
                 git_is_dirty=None):
        """
        Construct a new VersionInfo object containing the specified version
        information.

        :param pip_version: the package version reported by pip
        :type pip_version: str
        :param pip_url: pip package "Home-page" metadata value
        :type pip_url: str
        :param pip_requirement: the package requirement as installed via pip
        :type pip_requirement: str
        :param pkg_resources_version: the package version reported by
          pkg_resources
        :type pkg_resources_version: str
        :param pkg_resources_url: pkg_resources package "Home-page" metadata
          value
        :type pkg_resources_url: str
        :param git_tag: if the package source has a git repository on disk,
          the tag matching the current commit
        :type git_tag: str
        :param git_commit: if the package source has a git repository on disk,
          the commit SHA that repository is currently at
        :type git_commit: str
        :param git_remotes: if the package source has a git repository on disk,
          a dict of name to URL pairs for each git remote
        :type git_remotes: dict
        :param git_is_dirty: if the package source has a git repository on disk,
          whether or not that repository has uncommitted changes or is behind
          origin.
        :type git_is_dirty: bool
        """
        self._pip_version = pip_version
        self._pip_url = pip_url
        self._pip_requirement = pip_requirement
        self._pkg_resources_version = pkg_resources_version
        self._pkg_resources_url = pkg_resources_url
        self._git_tag = git_tag
        self._git_commit = git_commit
        self._git_remotes = git_remotes
        self._git_is_dirty = git_is_dirty

    @property
    def version(self):
        """
        Return the package/distribution version, from pip if possible or else
        from pkg_resources, or else None if neither can be found.

        :return: package/distribution version
        :rtype: :py:obj:`str` or :py:data:`None`
        """
        if self._pip_version is not None:
            return self._pip_version
        return self._pkg_resources_version

    @property
    def url(self):
        """
        Return the package/distribution "Home-page", from pip if possible or
        else from pkg_resources, or else None if neither can be found.

        :return: package/distribution Home-page/URL
        :rtype: :py:obj:`str` or :py:data:`None`
        """
        if self._pip_url is not None:
            return self._pip_url
        return self._pkg_resources_url

    @property
    def pip_version(self):
        """
        Return the pip distribution version, or None if the distribution cannot
        be found with pip.

        :return: pip distribution version
        :rtype: :py:obj:`str` or :py:data:`None`
        """
        return self._pip_version

    @property
    def pip_url(self):
        """
        Return the pip distribution "Home-page", or None if the distribution
        cannot be found with pip.

        :return: pip distribution Home-page/URL
        :rtype: :py:obj:`str` or :py:data:`None`
        """
        return self._pip_url

    @property
    def pip_requirement(self):
        """
        Return the pip requirement for the current installation of the
        distribution, or None if the distribution cannot be found with pip.

        :return: pip requirement string
        :rtype: :py:obj:`str` or :py:data:`None`
        """
        return self._pip_requirement

    @property
    def pkg_resources_version(self):
        """
        Return the pkg_resources distribution version, or None if the
        distribution cannot be found with pkg_resources.

        :return: pkg_resources distribution version
        :rtype: :py:obj:`str` or :py:data:`None`
        """
        return self._pkg_resources_version

    @property
    def pkg_resources_url(self):
        """
        Return the pkg_resources distribution "Home-page", or None if the
        distribution cannot be found with pkg_resources.

        :return: pkg_resources distribution Home-page/URL
        :rtype: :py:obj:`str` or :py:data:`None`
        """
        return self._pkg_resources_url

    @property
    def git_tag(self):
        """
        Return the name of the git that that the distribution is installed at,
        or None if there is no tag matching the current commit, or if not
        installed via git.

        :return: current git tag
        :rtype: :py:obj:`str` or :py:data:`None`
        """
        return self._git_tag

    @property
    def git_commit(self):
        """
        Return the hex SHA of the current git commit that the distribution is
        installed at, or None if not installed via git.

        :return: git commit hex SHA
        :rtype: :py:obj:`str` or :py:data:`None`
        """
        return self._git_commit

    @property
    def git_remotes(self):
        """
        If the distribution is installed via git, return a dict of all remotes
        configured on the git repository; keys are the remote name and values
        are the remote's first URL. If not installed via git, return None.

        :return: dict of git remotes, name (str) to first URL (str)
        :rtype: :py:obj:`dict` or :py:data:`None`
        """
        return self._git_remotes

    @property
    def git_remote(self):
        """
        If the distribution is installed via git, return the first URL of the
        'origin' remote if one is configured for the repo, or else the first
        URL of the lexicographically-first remote, or else None.

        :return: origin or first remote URL
        :rtype: :py:obj:`str` or :py:data:`None`
        """
        if self._git_remotes is None or len(self._git_remotes) < 1:
            return None
        if 'origin' in self._git_remotes:
            return self._git_remotes['origin']
        k = sorted(self._git_remotes.keys())[0]
        return self._git_remotes[k]

    @property
    def git_is_dirty(self):
        """
        Return True if the distribution is installed via git and has uncommitted
        changes or untracked files in the repo; Return False if the distribution
        is installed via git and does not have uncommitted changes or untracked
        files in the repo; return None if the distribution is not installed
        via git.

        :return: whether or not the git repo has uncommitted changes or
          untracked files
        :rtype: :py:obj:`bool` or :py:data:`None`
        """
        return self._git_is_dirty

    @property
    def git_str(self):
        """
        If the distribution is not installed via git, return an empty string.

        If the distribution is installed via git and pip recognizes the git
        source, return the pip requirement string specifying the git URL and
        commit, with an '*' appended if :py:attr:`~.git_is_dirty` is True.

        Otherwise, return a string of the form:

            url@ref[*]

        Where URL is the remote URL, ref is the tag name if the repo is checked
        out to a commit that matches a tag or else the commit hex SHA, and '*'
        is appended if :py:attr:`~.git_is_dirty` is True.

        :return: description of the git repo remote and state
        :rtype: str
        """
        dirty = '*' if self._git_is_dirty else ''
        if 'git' in self._pip_requirement:
            return self._pip_requirement + dirty
        if self._git_commit is None and self._git_remotes is None:
            return ''
        ref = self._git_tag if self._git_tag is not None else self._git_commit
        return '%s@%s%s' % (self.git_remote, ref, dirty)

    @property
    def short_str(self):
        """
        Return a string of the form "ver <url>" where ver is the distribution
        version and URL is the distribution Home-Page url, or '' if neither
        can be found.

        :return: version and URL
        :rtype: str
        """
        if self.version is None and self.url is None:
            return ''
        return '%s <%s>' % (self.version, self.url)

    @property
    def long_str(self):
        """
        Return a long version and installation specifier string of the form:

        If :py:attr:`~.git_str` == '':

            SHORT_STR

        otherwise:

            SHORT_STR (GIT_STR)

        Where ``SHORT_STR`` is :py:attr:`~.short_str` and ``GIT_STR`` is
        :py:attr:`~.git_str`.

        :return: long version/installation specifier string
        :rtype: str
        """
        gs = self.git_str
        if gs == '':
            return self.short_str
        return self.short_str + ' (' + gs + ')'

    @property
    def as_dict(self):
        """
        Return the constructor arguments as a dictionary (effectively the
        kwargs to the constructor).

        :return: dict of constructor arguments
        :rtype: dict
        """
        return {
            'pip_version': self._pip_version,
            'pip_url': self._pip_url,
            'pip_requirement': self._pip_requirement,
            'pkg_resources_version': self._pkg_resources_version,
            'pkg_resources_url': self._pkg_resources_url,
            'git_tag': self._git_tag,
            'git_commit': self._git_commit,
            'git_remotes': self._git_remotes,
            'git_is_dirty': self._git_is_dirty,
        }

    def __repr__(self):
        """
        Return a string representation of the object.

        :return: representation of the object
        :rtype: str
        """
        d = self.as_dict
        p = ', '.join(['%s=%s' % (k, d[k]) for k in sorted(d.keys())])
        return 'VersionInfo(%s)' % p

    def __eq__(self, other):
        """

        :param other: class to compare
        :type other: VersionInfo
        :return: whether or not the class' as_dict properties are equal
        :rtype: bool
        """
        return self.as_dict == other.as_dict
